@@ -19,12 +19,19 @@
 
 #include "grn.h"
 
+#include "grn_bulk.h"
 #include "grn_db.h"
 #include "grn_expr.h"
+#include "grn_expr_executor.h"
 #include "grn_group.h"
+#include "grn_output_columns.h"
 #include "grn_table.h"
 #include "grn_vector.h"
 
+#ifdef GRN_WITH_XXHASH
+# include <xxhash.h>
+#endif
+
 #define GRN_TABLE_GROUP_BY_KEY           0
 #define GRN_TABLE_GROUP_BY_VALUE         1
 #define GRN_TABLE_GROUP_BY_COLUMN_VALUE  2
@@ -32,6 +39,41 @@
 #define GRN_TABLE_GROUP_FILTER_PREFIX    0
 #define GRN_TABLE_GROUP_FILTER_SUFFIX    (1L<<2)
 
+#ifdef GRN_WITH_XXHASH
+static bool grn_table_group_fallback_hash_value_enable = false;
+static bool grn_table_group_all_hash_value_enable = false;
+#endif
+
+void
+grn_group_init_from_env(void)
+{
+#ifdef GRN_WITH_XXHASH
+  {
+    char grn_table_group_fallback_hash_value_enable_env[GRN_ENV_BUFFER_SIZE];
+    grn_getenv("GRN_TABLE_GROUP_FALLBACK_HASH_VALUE_ENABLE",
+               grn_table_group_fallback_hash_value_enable_env,
+               GRN_ENV_BUFFER_SIZE);
+    if (strcmp(grn_table_group_fallback_hash_value_enable_env, "no") == 0) {
+      grn_table_group_fallback_hash_value_enable = false;
+    } else {
+      grn_table_group_fallback_hash_value_enable = true;
+    }
+  }
+
+  {
+    char grn_table_group_all_hash_value_enable_env[GRN_ENV_BUFFER_SIZE];
+    grn_getenv("GRN_TABLE_GROUP_ALL_HASH_VALUE_ENABLE",
+               grn_table_group_all_hash_value_enable_env,
+               GRN_ENV_BUFFER_SIZE);
+    if (strcmp(grn_table_group_all_hash_value_enable_env, "yes") == 0) {
+      grn_table_group_all_hash_value_enable = true;
+    } else {
+      grn_table_group_all_hash_value_enable = false;
+    }
+  }
+#endif
+}
+
 grn_rc
 grn_table_group_aggregator_init(grn_ctx *ctx,
                                 grn_table_group_aggregator *aggregator)
@@ -562,17 +604,64 @@ grn_table_group_add_subrec(grn_ctx *ctx,
   }
 }
 
+typedef struct {
+  grn_obj *object;
+  grn_obj *range;
+  bool is_expr;
+  grn_expr_executor expr_executor;
+  bool is_reference_column;
+} group_key;
+
+static grn_inline void
+group_key_init(grn_ctx *ctx,
+               group_key *key,
+               grn_obj *object)
+{
+  key->object = object;
+  grn_id range_id = grn_obj_get_range(ctx, key->object);
+  key->range = grn_ctx_at(ctx, range_id);
+  key->is_expr = grn_obj_is_expr(ctx, key->object);
+  if (key->is_expr) {
+    grn_expr_executor_init(ctx, &(key->expr_executor), key->object);
+  }
+  key->is_reference_column = grn_obj_is_table(ctx, key->range);
+}
+
+static grn_inline void
+group_key_fin(grn_ctx *ctx,
+              group_key *key)
+{
+  if (key->is_expr) {
+    grn_expr_executor_fin(ctx, &(key->expr_executor));
+  }
+  grn_obj_unref(ctx, key->range);
+}
+
+static grn_inline grn_obj *
+group_key_get_value(grn_ctx *ctx,
+                    group_key *key,
+                    grn_id id,
+                    grn_obj *buffer)
+{
+  if (key->is_expr) {
+    return grn_expr_executor_exec(ctx, &(key->expr_executor), id);
+  } else {
+    GRN_BULK_REWIND(buffer);
+    grn_obj_get_value(ctx, key->object, id, buffer);
+    return buffer;
+  }
+}
+
 typedef struct {
   grn_obj bulk;
   grn_obj value_buffer;
   bool need_resolved_id;
   grn_obj *table;
-  grn_obj *key;
+  group_key key;
   grn_obj *range;
   grn_ra_cache cache;
   grn_obj *res;
   bool with_subrec;
-  bool is_reference_column;
 } single_key_records_data;
 
 static grn_inline grn_id
@@ -597,7 +686,7 @@ grn_table_group_single_key_records_foreach_fix_size(grn_ctx *ctx,
                                                     void *user_data)
 {
   single_key_records_data *data = user_data;
-  grn_ra *ra = (grn_ra *)(data->key);
+  grn_ra *ra = (grn_ra *)(data->key.object);
   grn_rset_recinfo *ri = NULL;
 
   if (data->with_subrec) {
@@ -606,12 +695,12 @@ grn_table_group_single_key_records_foreach_fix_size(grn_ctx *ctx,
   grn_id value_id =
     grn_table_group_single_key_records_resolve_id(ctx, data, cursor, id);
   void *column_value = grn_ra_ref_cache(ctx, ra, value_id, &(data->cache));
-  if (data->is_reference_column) {
+  if (data->key.is_reference_column) {
     grn_id reference_id = *((grn_id *)column_value);
     if (reference_id == GRN_ID_NIL) {
       return ctx->rc;
     }
-    if (grn_table_at(ctx, data->range, reference_id) == GRN_ID_NIL) {
+    if (grn_table_at(ctx, data->key.range, reference_id) == GRN_ID_NIL) {
       return ctx->rc;
     }
   }
@@ -654,7 +743,7 @@ grn_table_group_single_key_records_foreach_var_size_reference(
   void *user_data)
 {
   single_key_records_data *data = user_data;
-  grn_ja *ja = (grn_ja *)(data->key);
+  grn_ja *ja = (grn_ja *)(data->key.object);
   grn_rset_recinfo *ri = NULL;
 
   if (data->with_subrec) {
@@ -713,19 +802,18 @@ grn_table_group_single_key_records_foreach(grn_ctx *ctx,
                                            void *user_data)
 {
   single_key_records_data *data = user_data;
-  grn_obj *bulk = &(data->bulk);
   grn_obj *value_buffer = &(data->value_buffer);
   void *value;
   grn_rset_recinfo *ri = NULL;
 
-  GRN_BULK_REWIND(bulk);
   if (data->with_subrec) {
     grn_table_cursor_get_value(ctx, cursor, (void **)&ri);
   }
   grn_id value_id =
     grn_table_group_single_key_records_resolve_id(ctx, data, cursor, id);
-  grn_obj_get_value(ctx, data->key, value_id, bulk);
-  switch (data->bulk.header.type) {
+  grn_obj *bulk =
+    group_key_get_value(ctx, &(data->key), value_id, &(data->bulk));
+  switch (bulk->header.type) {
   case GRN_UVECTOR :
     {
       unsigned int element_size;
@@ -738,7 +826,7 @@ grn_table_group_single_key_records_foreach(grn_ctx *ctx,
       for (i = 0; i < n_elements; i++) {
         uint8_t *element = elements + (element_size * i);
 
-        if (data->is_reference_column) {
+        if (data->key.is_reference_column) {
           grn_id reference_id = *((grn_id *)element);
           if (reference_id == GRN_ID_NIL) {
             return ctx->rc;
@@ -783,7 +871,7 @@ grn_table_group_single_key_records_foreach(grn_ctx *ctx,
   case GRN_BULK :
     {
       grn_id group_id;
-      if ((!data->is_reference_column ||
+      if ((!data->key.is_reference_column ||
            *((grn_id *)GRN_BULK_HEAD(bulk))) &&
           (group_id = grn_table_add_v(ctx, data->res,
                                       GRN_BULK_HEAD(bulk),
@@ -813,33 +901,31 @@ grn_table_group_single_key_records(grn_ctx *ctx, grn_obj *table,
   GRN_TEXT_INIT(&(data.bulk), 0);
   GRN_VOID_INIT(&(data.value_buffer));
   data.table = table;
-  data.key = key;
   data.res = result->table;
   if ((cursor = grn_table_cursor_open(ctx, table, NULL, 0, NULL, 0, 0, -1, 0))) {
+    grn_obj *real_key = key;
     data.need_resolved_id = false;
     if (key->header.type == GRN_ACCESSOR) {
       grn_accessor *a = (grn_accessor *)key;
       if (a->action == GRN_ACCESSOR_GET_KEY &&
           a->next && a->next->action == GRN_ACCESSOR_GET_COLUMN_VALUE &&
           a->next->obj && !a->next->next) {
-        data.key = a->next->obj;
+        real_key = a->next->obj;
         data.need_resolved_id = true;
       }
     }
+    group_key_init(ctx, &(data.key), real_key);
     data.with_subrec = (DB_OBJ(table)->header.flags & GRN_OBJ_WITH_SUBREC);
-    grn_id range_id = grn_obj_get_range(ctx, data.key);
-    data.range = grn_ctx_at(ctx, range_id);
-    data.is_reference_column = grn_obj_is_table(ctx, data.range);
-    if (data.key->header.type == GRN_COLUMN_FIX_SIZE) {
-      GRN_RA_CACHE_INIT((grn_ra *)(data.key), &(data.cache));
+    if (data.key.object->header.type == GRN_COLUMN_FIX_SIZE) {
+      GRN_RA_CACHE_INIT((grn_ra *)(data.key.object), &(data.cache));
       grn_table_cursor_foreach(
         ctx,
         cursor,
         grn_table_group_single_key_records_foreach_fix_size,
         &data);
-      GRN_RA_CACHE_FIN((grn_ra *)(data.key), &(data.cache));
-    } else if (data.key->header.type == GRN_COLUMN_VAR_SIZE &&
-               data.is_reference_column) {
+      GRN_RA_CACHE_FIN((grn_ra *)(data.key.object), &(data.cache));
+    } else if (data.key.object->header.type == GRN_COLUMN_VAR_SIZE &&
+               data.key.is_reference_column) {
       grn_table_cursor_foreach(
         ctx,
         cursor,
@@ -851,7 +937,7 @@ grn_table_group_single_key_records(grn_ctx *ctx, grn_obj *table,
                                grn_table_group_single_key_records_foreach,
                                &data);
     }
-    grn_obj_unref(ctx, data.range);
+    group_key_fin(ctx, &(data.key));
     grn_table_cursor_close(ctx, cursor);
   }
   GRN_OBJ_FIN(ctx, &(data.value_buffer));
@@ -1004,125 +1090,158 @@ grn_table_group_with_range_gap(grn_ctx *ctx, grn_obj *table,
   return 0;
 }
 
+typedef struct {
+  grn_obj *table;
+  group_key *keys;
+  int n_keys;
+  grn_table_group_result *results;
+  int n_results;
+  grn_obj bulk;
+  grn_obj vector;
+  grn_obj vector_pack_header;
+  grn_obj vector_pack_footer;
+  grn_id id;
+  grn_rset_recinfo *ri;
+} grn_table_group_multi_keys_data;
+
 static grn_inline void
 grn_table_group_multi_keys_add_record(grn_ctx *ctx,
-                                      grn_table_sort_key *keys,
-                                      int n_keys,
-                                      grn_table_group_result *results,
-                                      int n_results,
-                                      grn_id id,
-                                      grn_rset_recinfo *ri,
-                                      grn_obj *vector,
-                                      grn_obj *bulk)
+                                      grn_table_group_multi_keys_data *data)
 {
-  int r;
-  grn_table_group_result *rp;
+  grn_obj *bulk = &(data->bulk);
+  grn_obj *vector = &(data->vector);
+  grn_obj *vector_pack_header = &(data->vector_pack_header);
+  grn_obj *vector_pack_footer = &(data->vector_pack_footer);
+
+  int i;
+  for (i = 0; i < data->n_results; i++) {
+    grn_table_group_result *result = data->results + i;
 
-  grn_obj header;
-  grn_obj footer;
-  GRN_TEXT_INIT(&header, 0);
-  GRN_TEXT_INIT(&footer, 0);
-  for (r = 0, rp = results; r < n_results; r++, rp++) {
-    void *value;
     int end;
-
-    if (rp->key_end > n_keys) {
-      end = n_keys;
+    if (result->key_end > data->n_keys) {
+      end = data->n_keys;
     } else {
-      end = rp->key_end + 1;
+      end = result->key_end + 1;
     }
+    GRN_BULK_REWIND(vector_pack_header);
+    GRN_BULK_REWIND(vector_pack_footer);
     grn_obj *body = grn_vector_pack(ctx,
                                     vector,
-                                    rp->key_begin,
-                                    end - rp->key_begin,
+                                    result->key_begin,
+                                    end - result->key_begin,
                                     0,
-                                    &header,
-                                    &footer);
+                                    vector_pack_header,
+                                    vector_pack_footer);
     GRN_BULK_REWIND(bulk);
-    GRN_TEXT_PUT(ctx, bulk, GRN_TEXT_VALUE(&header), GRN_TEXT_LEN(&header));
+#ifdef GRN_WITH_XXHASH
+    bool need_hash_value = false;
+    if (grn_table_group_all_hash_value_enable) {
+      need_hash_value = true;
+    } else if (grn_table_group_fallback_hash_value_enable) {
+      const size_t total_size =
+        GRN_TEXT_LEN(vector_pack_header) +
+        GRN_TEXT_LEN(body) +
+        GRN_TEXT_LEN(vector_pack_footer);
+      if (total_size > GRN_TABLE_MAX_KEY_SIZE) {
+        need_hash_value = true;
+      }
+    }
+#endif
+    GRN_TEXT_PUT(ctx,
+                 bulk,
+                 GRN_TEXT_VALUE(vector_pack_header),
+                 GRN_TEXT_LEN(vector_pack_header));
     GRN_TEXT_PUT(ctx, bulk, GRN_TEXT_VALUE(body), GRN_TEXT_LEN(body));
-    GRN_TEXT_PUT(ctx, bulk, GRN_TEXT_VALUE(&footer), GRN_TEXT_LEN(&footer));
+    GRN_TEXT_PUT(ctx,
+                 bulk,
+                 GRN_TEXT_VALUE(vector_pack_footer),
+                 GRN_TEXT_LEN(vector_pack_footer));
+#ifdef GRN_WITH_XXHASH
+    if (need_hash_value) {
+      XXH128_hash_t hash_value = XXH3_128bits(GRN_TEXT_VALUE(bulk),
+                                              GRN_TEXT_LEN(bulk));
+      GRN_BULK_REWIND(bulk);
+      /* This value must not be conflicted with the first byte of data
+       * packed by grn_vector_pack(). The first byte is generated by
+       * grn_text_benc(ctx, header, n) where n is the number of
+       * drilldown keys. n must be less than 0x1020408f. See
+       * GRN_B_ENC() for 0x1020408f. */
+      const uint8_t hash_value_mark = 0x8f;
+      GRN_UINT8_PUT(ctx, bulk, hash_value_mark);
+      GRN_UINT64_PUT(ctx, bulk, hash_value.low64);
+      GRN_UINT64_PUT(ctx, bulk, hash_value.high64);
+    }
+#endif
 
-    // todo : cut off GRN_ID_NIL
-    grn_id group_id = grn_table_add_v(ctx, rp->table,
-                                      GRN_BULK_HEAD(bulk), GRN_BULK_VSIZE(bulk),
-                                      &value, NULL);
+    void *value;
+    grn_id group_id = grn_table_add_v(ctx,
+                                      result->table,
+                                      GRN_BULK_HEAD(bulk),
+                                      GRN_BULK_VSIZE(bulk),
+                                      &value,
+                                      NULL);
     if (group_id != GRN_ID_NIL) {
-      grn_table_group_add_subrec(ctx, rp->table, group_id, value,
-                                 ri ? ri->score : 0,
-                                 (grn_rset_posinfo *)&id, 0,
+      grn_table_group_add_subrec(ctx,
+                                 result->table,
+                                 group_id,
+                                 value,
+                                 data->ri ? data->ri->score : 0,
+                                 (grn_rset_posinfo *)&(data->id),
+                                 0,
                                  bulk);
     }
   }
-  GRN_OBJ_FIN(ctx, &header);
-  GRN_OBJ_FIN(ctx, &footer);
 }
 
 static void
 grn_table_group_multi_keys_scalar_records(grn_ctx *ctx,
-                                          grn_obj *table,
-                                          grn_table_sort_key *keys,
-                                          int n_keys,
-                                          grn_table_group_result *results,
-                                          int n_results)
+                                          grn_table_group_multi_keys_data *data)
 {
-  grn_id id;
-  grn_table_cursor *tc;
-  grn_obj bulk;
-  grn_obj vector;
-
-  tc = grn_table_cursor_open(ctx, table, NULL, 0, NULL, 0, 0, -1, 0);
+  grn_table_cursor *tc = grn_table_cursor_open(ctx,
+                                               data->table,
+                                               NULL, 0,
+                                               NULL, 0,
+                                               0, -1, 0);
   if (!tc) {
     return;
   }
 
-  GRN_TEXT_INIT(&bulk, 0);
-  GRN_OBJ_INIT(&vector, GRN_VECTOR, 0, GRN_DB_VOID);
-  while ((id = grn_table_cursor_next(ctx, tc))) {
-    int k;
-    grn_table_sort_key *kp;
-    grn_rset_recinfo *ri = NULL;
-
-    if (DB_OBJ(table)->header.flags & GRN_OBJ_WITH_SUBREC) {
-      grn_table_cursor_get_value(ctx, tc, (void **)&ri);
+  grn_obj *bulk = &(data->bulk);
+  grn_obj *vector = &(data->vector);
+  const grn_obj_flags flags = DB_OBJ(data->table)->header.flags;
+  const bool with_subrec = (flags & GRN_OBJ_WITH_SUBREC);
+  while ((data->id = grn_table_cursor_next(ctx, tc))) {
+    if (with_subrec) {
+      grn_table_cursor_get_value(ctx, tc, (void **)&(data->ri));
     }
 
-    GRN_BULK_REWIND(&vector);
-    for (k = 0, kp = keys; k < n_keys; k++, kp++) {
-      GRN_BULK_REWIND(&bulk);
-      grn_obj_get_value(ctx, kp->key, id, &bulk);
-      grn_vector_add_element(ctx, &vector,
-                             GRN_BULK_HEAD(&bulk), GRN_BULK_VSIZE(&bulk),
+    GRN_BULK_REWIND(vector);
+    int i;
+    for (i = 0; i < data->n_keys; i++) {
+      grn_obj *value = group_key_get_value(ctx, data->keys + i, data->id, bulk);
+      grn_vector_add_element(ctx, vector,
+                             GRN_BULK_HEAD(value),
+                             GRN_BULK_VSIZE(value),
                              0,
-                             bulk.header.domain);
+                             value->header.domain);
     }
 
-    grn_table_group_multi_keys_add_record(ctx, keys, n_keys, results, n_results,
-                                          id, ri, &vector, &bulk);
+    grn_table_group_multi_keys_add_record(ctx, data);
   }
-  GRN_OBJ_FIN(ctx, &vector);
-  GRN_OBJ_FIN(ctx, &bulk);
   grn_table_cursor_close(ctx, tc);
 }
 
 static grn_inline void
 grn_table_group_multi_keys_vector_record(grn_ctx *ctx,
-                                         grn_table_sort_key *keys,
+                                         grn_table_group_multi_keys_data *data,
                                          grn_obj *key_buffers,
-                                         int nth_key,
-                                         int n_keys,
-                                         grn_table_group_result *results,
-                                         int n_results,
-                                         grn_id id,
-                                         grn_rset_recinfo *ri,
-                                         grn_obj *vector,
-                                         grn_obj *bulk)
+                                         int nth_key)
 {
-  int k;
-  grn_table_sort_key *kp;
+  grn_obj *vector = &(data->vector);
 
-  for (k = nth_key, kp = &(keys[nth_key]); k < n_keys; k++, kp++) {
-    grn_obj *key_buffer = &(key_buffers[k]);
+  int i;
+  for (i = nth_key; i < data->n_keys; i++) {
+    grn_obj *key_buffer = &(key_buffers[i]);
     switch (key_buffer->header.type) {
     case GRN_UVECTOR :
       {
@@ -1130,24 +1249,23 @@ grn_table_group_multi_keys_vector_record(grn_ctx *ctx,
         unsigned int element_size;
         grn_id domain;
         uint8_t *elements;
-        unsigned int i, n_elements;
+        unsigned int j, n_elements;
 
         n_vector_elements = grn_vector_size(ctx, vector);
         domain = key_buffer->header.domain;
         elements = GRN_BULK_HEAD(key_buffer);
         element_size = grn_uvector_element_size(ctx, key_buffer);
         n_elements = GRN_BULK_VSIZE(key_buffer) / element_size;
-        for (i = 0; i < n_elements; i++) {
-          const char *element_id = elements + (element_size * i);
+        for (j = 0; j < n_elements; j++) {
+          const char *element_id = elements + (element_size * j);
           grn_vector_add_element(ctx, vector,
                                  element_id, sizeof(grn_id),
                                  0,
                                  domain);
           grn_table_group_multi_keys_vector_record(ctx,
-                                                   keys, key_buffers,
-                                                   k + 1, n_keys,
-                                                   results, n_results,
-                                                   id, ri, vector, bulk);
+                                                   data,
+                                                   key_buffers,
+                                                   i + 1);
           while (grn_vector_size(ctx, vector) != n_vector_elements) {
             const char *content;
             grn_vector_pop_element(ctx, vector, &content, NULL, NULL);
@@ -1159,25 +1277,24 @@ grn_table_group_multi_keys_vector_record(grn_ctx *ctx,
     case GRN_VECTOR :
       {
         unsigned int n_vector_elements;
-        unsigned int i, n_key_elements;
+        unsigned int j, n_key_elements;
 
         n_vector_elements = grn_vector_size(ctx, vector);
         n_key_elements = grn_vector_size(ctx, key_buffer);
-        for (i = 0; i < n_key_elements; i++) {
+        for (j = 0; j < n_key_elements; j++) {
           const char *content;
           unsigned int content_length;
           grn_id domain;
-          content_length = grn_vector_get_element(ctx, key_buffer, i,
+          content_length = grn_vector_get_element(ctx, key_buffer, j,
                                                   &content, NULL, &domain);
           grn_vector_add_element(ctx, vector,
                                  content, content_length,
                                  0,
                                  domain);
           grn_table_group_multi_keys_vector_record(ctx,
-                                                   keys, key_buffers,
-                                                   k + 1, n_keys,
-                                                   results, n_results,
-                                                   id, ri, vector, bulk);
+                                                   data,
+                                                   key_buffers,
+                                                   i + 1);
           while (grn_vector_size(ctx, vector) != n_vector_elements) {
             grn_vector_pop_element(ctx, vector, &content, NULL, NULL);
           }
@@ -1194,71 +1311,58 @@ grn_table_group_multi_keys_vector_record(grn_ctx *ctx,
     }
   }
 
-  if (k == n_keys) {
-    grn_table_group_multi_keys_add_record(ctx,
-                                          keys, n_keys,
-                                          results, n_results,
-                                          id, ri, vector, bulk);
+  if (i == data->n_keys) {
+    grn_table_group_multi_keys_add_record(ctx, data);
   }
 }
 
 static void
 grn_table_group_multi_keys_vector_records(grn_ctx *ctx,
-                                          grn_obj *table,
-                                          grn_table_sort_key *keys,
-                                          int n_keys,
-                                          grn_table_group_result *results,
-                                          int n_results)
+                                          grn_table_group_multi_keys_data *data)
 {
-  grn_id id;
-  grn_table_cursor *tc;
-  grn_obj bulk;
-  grn_obj vector;
-  grn_obj *key_buffers;
-  int k;
-
-  tc = grn_table_cursor_open(ctx, table, NULL, 0, NULL, 0, 0, -1, 0);
+  grn_table_cursor *tc = grn_table_cursor_open(ctx,
+                                               data->table,
+                                               NULL, 0,
+                                               NULL, 0,
+                                               0, -1, 0);
   if (!tc) {
     return;
   }
 
-  key_buffers = GRN_MALLOCN(grn_obj, n_keys);
+  grn_obj *key_buffers = GRN_MALLOCN(grn_obj, data->n_keys);
   if (!key_buffers) {
     grn_table_cursor_close(ctx, tc);
     return;
   }
-
-  GRN_TEXT_INIT(&bulk, 0);
-  GRN_OBJ_INIT(&vector, GRN_VECTOR, 0, GRN_DB_VOID);
-  for (k = 0; k < n_keys; k++) {
-    GRN_VOID_INIT(&(key_buffers[k]));
+  int i;
+  for (i = 0; i < data->n_keys; i++) {
+    GRN_VOID_INIT(&(key_buffers[i]));
   }
-  while ((id = grn_table_cursor_next(ctx, tc))) {
-    grn_table_sort_key *kp;
-    grn_rset_recinfo *ri = NULL;
 
-    if (DB_OBJ(table)->header.flags & GRN_OBJ_WITH_SUBREC) {
-      grn_table_cursor_get_value(ctx, tc, (void **)&ri);
+  grn_obj *vector = &(data->vector);
+  const grn_obj_flags flags = (DB_OBJ(data->table)->header.flags);
+  const bool with_subrec = (flags & GRN_OBJ_WITH_SUBREC);
+  while ((data->id = grn_table_cursor_next(ctx, tc))) {
+    if (with_subrec) {
+      grn_table_cursor_get_value(ctx, tc, (void **)&(data->ri));
     }
 
-    for (k = 0, kp = keys; k < n_keys; k++, kp++) {
-      grn_obj *key_buffer = &(key_buffers[k]);
-      GRN_BULK_REWIND(key_buffer);
-      grn_obj_get_value(ctx, kp->key, id, key_buffer);
+    for (i = 0; i < data->n_keys; i++) {
+      group_key *key = data->keys + i;
+      grn_obj *key_buffer = &(key_buffers[i]);
+      grn_obj *value = group_key_get_value(ctx, key, data->id, key_buffer);
+      if (value != key_buffer) {
+        grn_bulk_copy(ctx, value, key_buffer);
+      }
     }
 
-    GRN_BULK_REWIND(&vector);
-    grn_table_group_multi_keys_vector_record(ctx,
-                                             keys, key_buffers, 0, n_keys,
-                                             results, n_results,
-                                             id, ri, &vector, &bulk);
+    GRN_BULK_REWIND(vector);
+    grn_table_group_multi_keys_vector_record(ctx, data, key_buffers, 0);
   }
-  for (k = 0; k < n_keys; k++) {
-    GRN_OBJ_FIN(ctx, &(key_buffers[k]));
+  for (i = 0; i < data->n_keys; i++) {
+    GRN_OBJ_FIN(ctx, &(key_buffers[i]));
   }
   GRN_FREE(key_buffers);
-  GRN_OBJ_FIN(ctx, &vector);
-  GRN_OBJ_FIN(ctx, &bulk);
   grn_table_cursor_close(ctx, tc);
 }
 
@@ -1286,11 +1390,17 @@ grn_table_group(grn_ctx *ctx, grn_obj *table,
                 grn_table_group_result *results, int n_results)
 {
   grn_rc rc = GRN_SUCCESS;
-  grn_bool group_by_all_records = GRN_FALSE;
+  bool group_by_all_records = false;
   if (n_keys == 0 && n_results == 1) {
     group_by_all_records = GRN_TRUE;
-  } else if (!table || !n_keys || !n_results) {
-    ERR(GRN_INVALID_ARGUMENT, "table or n_keys or n_results is void");
+  } else if (!table) {
+    ERR(GRN_INVALID_ARGUMENT, "[table][group] table is NULL");
+    return GRN_INVALID_ARGUMENT;
+  } else if (n_keys == 0) {
+    ERR(GRN_INVALID_ARGUMENT, "[table][group] no keys");
+    return GRN_INVALID_ARGUMENT;
+  } else if (n_results == 0) {
+    ERR(GRN_INVALID_ARGUMENT, "[table][group] no results");
     return GRN_INVALID_ARGUMENT;
   }
   GRN_API_ENTER;
@@ -1300,7 +1410,7 @@ grn_table_group(grn_ctx *ctx, grn_obj *table,
     grn_table_group_result *rp;
     for (k = 0, kp = keys; k < n_keys; k++, kp++) {
       if ((kp->flags & GRN_TABLE_GROUP_BY_COLUMN_VALUE) && !kp->key) {
-        ERR(GRN_INVALID_ARGUMENT, "column missing in (%d)", k);
+        ERR(GRN_INVALID_ARGUMENT, "[table][group] column is missing: %d", k);
         goto exit;
       }
     }
@@ -1316,7 +1426,19 @@ grn_table_group(grn_ctx *ctx, grn_obj *table,
         if (group_by_all_records) {
           key_type = grn_ctx_at(ctx, GRN_DB_SHORT_TEXT);
         } else if (n_keys == 1) {
-          key_type = grn_ctx_at(ctx, grn_obj_get_range(ctx, keys[0].key));
+          grn_id key_type_id = grn_obj_get_range(ctx, keys[0].key);
+          if (key_type_id == GRN_ID_NIL) {
+            grn_obj inspected;
+            GRN_TEXT_INIT(&inspected, 0);
+            grn_inspect(ctx, &inspected, keys[0].key);
+            ERR(GRN_INVALID_ARGUMENT,
+                "[table][group] unknown key type: %.*s",
+                (int)GRN_TEXT_LEN(&inspected),
+                GRN_TEXT_VALUE(&inspected));
+            GRN_OBJ_FIN(ctx, &inspected);
+            goto exit;
+          }
+          key_type = grn_ctx_at(ctx, key_type_id);
         } else {
           flags |= GRN_OBJ_KEY_VAR_SIZE;
         }
@@ -1327,6 +1449,9 @@ grn_table_group(grn_ctx *ctx, grn_obj *table,
                                                         key_type, table,
                                                         rp->max_n_subrecs,
                                                         additional_value_size);
+        if (key_type) {
+          grn_obj_unref(ctx, key_type);
+        }
         if (!rp->table) {
           goto exit;
         }
@@ -1363,20 +1488,57 @@ grn_table_group(grn_ctx *ctx, grn_obj *table,
         grn_id range_id;
         grn_obj_flags range_flags = 0;
         grn_obj_get_range_info(ctx, kp->key, &range_id, &range_flags);
+        if (range_id == GRN_ID_NIL) {
+          grn_obj inspected;
+          GRN_TEXT_INIT(&inspected, 0);
+          grn_inspect(ctx, &inspected, kp->key);
+          ERR(GRN_INVALID_ARGUMENT,
+              "[table][group] unknown key type: %.*s",
+              (int)GRN_TEXT_LEN(&inspected),
+              GRN_TEXT_VALUE(&inspected));
+          GRN_OBJ_FIN(ctx, &inspected);
+          goto exit;
+        }
         if (range_flags == GRN_OBJ_VECTOR) {
           have_vector = GRN_TRUE;
           break;
         }
       }
-      if (have_vector) {
-        grn_table_group_multi_keys_vector_records(ctx, table,
-                                                  keys, n_keys,
-                                                  results, n_results);
-      } else {
-        grn_table_group_multi_keys_scalar_records(ctx, table,
-                                                  keys, n_keys,
-                                                  results, n_results);
+      grn_table_group_multi_keys_data data;
+      data.table = table;
+      data.keys = GRN_MALLOCN(group_key, n_keys);
+      if (!data.keys) {
+        ERR(GRN_NO_MEMORY_AVAILABLE,
+            "[table][group] failed to allocate group_keys: %s",
+            ctx->errbuf);
+        goto exit;
       }
+      data.n_keys = n_keys;
+      int i;
+      for (i = 0; i < n_keys; i++) {
+        group_key_init(ctx, data.keys + i, keys[i].key);
+      }
+      data.results = results;
+      data.n_results = n_results;
+      GRN_TEXT_INIT(&(data.bulk), 0);
+      GRN_OBJ_INIT(&(data.vector), GRN_VECTOR, 0, GRN_DB_VOID);
+      GRN_TEXT_INIT(&(data.vector_pack_header), 0);
+      GRN_TEXT_INIT(&(data.vector_pack_footer), 0);
+      data.id = GRN_ID_NIL;
+      data.ri = NULL;
+      if (have_vector) {
+        grn_table_group_multi_keys_vector_records(ctx, &data);
+      } else {
+        grn_table_group_multi_keys_scalar_records(ctx, &data);
+      }
+      GRN_OBJ_FIN(ctx, &(data.vector_pack_footer));
+      GRN_OBJ_FIN(ctx, &(data.vector_pack_header));
+      GRN_OBJ_FIN(ctx, &(data.vector));
+      GRN_OBJ_FIN(ctx, &(data.bulk));
+      for (i = 0; i < n_keys; i++) {
+        group_key_fin(ctx, data.keys + i);
+      }
+      GRN_FREE(data.keys);
     }
     for (r = 0, rp = results; r < n_results; r++, rp++) {
       GRN_TABLE_GROUPED_ON(rp->table);
@@ -1394,3 +1556,121 @@ grn_table_group(grn_ctx *ctx, grn_obj *table,
 exit :
   GRN_API_RETURN(rc);
 }
+
+static void
+grn_table_group_keys_parse_one(grn_ctx *ctx,
+                               grn_obj *table,
+                               grn_obj *keys_expression,
+                               uint32_t code_start_offset,
+                               uint32_t code_end_offset,
+                               grn_table_sort_key *key)
+{
+  grn_expr *expr = (grn_expr *)keys_expression;
+  if ((code_end_offset - code_start_offset) == 1 &&
+      expr->codes[code_start_offset].op == GRN_OP_GET_VALUE &&
+      expr->codes[code_start_offset].value) {
+    key->key = expr->codes[code_start_offset].value;
+    grn_obj_refer(ctx, key->key);
+  } else {
+    grn_obj *sliced_expr = grn_expr_slice(ctx,
+                                          keys_expression,
+                                          code_start_offset,
+                                          code_end_offset);
+    key->key = sliced_expr;
+  }
+}
+
+grn_table_sort_key *
+grn_table_group_keys_parse(grn_ctx *ctx,
+                           grn_obj *table,
+                           const char *raw_keys,
+                           int32_t raw_keys_length,
+                           int *n_keys)
+{
+  GRN_API_ENTER;
+  if (raw_keys_length < 0) {
+    raw_keys_length = strlen(raw_keys);
+  }
+  grn_table_sort_key *keys = NULL;
+  if (grn_expr_is_v1_columns_format(ctx, raw_keys, raw_keys_length)) {
+    keys = grn_table_sort_key_from_str(ctx,
+                                       raw_keys,
+                                       raw_keys_length,
+                                       table,
+                                       n_keys);
+    GRN_API_RETURN(keys);
+  }
+
+  grn_obj offsets;
+  GRN_UINT32_INIT(&offsets, GRN_OBJ_VECTOR);
+  grn_obj *keys_expression = NULL;
+  grn_obj *variable;
+  GRN_EXPR_CREATE_FOR_QUERY(ctx, table, keys_expression, variable);
+  if (ctx->rc != GRN_SUCCESS) {
+    goto exit;
+  }
+  grn_expr_parse(ctx,
+                 keys_expression,
+                 raw_keys,
+                 raw_keys_length,
+                 NULL,
+                 GRN_OP_MATCH,
+                 GRN_OP_AND,
+                 GRN_EXPR_SYNTAX_GROUP_KEYS);
+  if (ctx->rc != GRN_SUCCESS) {
+    goto exit;
+  }
+
+  grn_output_columns_get_offsets(ctx, keys_expression, &offsets);
+  if (ctx->rc != GRN_SUCCESS) {
+    goto exit;
+  }
+
+  size_t n_offsets = GRN_UINT32_VECTOR_SIZE(&offsets) / 2;
+  keys = GRN_MALLOCN(grn_table_sort_key, n_offsets);
+  if (!keys) {
+    goto exit;
+  }
+  size_t i;
+  for (i = 0; i < n_offsets; i++) {
+    keys[i].key = NULL;
+  }
+
+  for (i = 0; i < n_offsets; i++) {
+    uint32_t code_start_offset;
+    uint32_t code_end_offset;
+    code_start_offset = GRN_UINT32_VALUE_AT(&offsets, i * 2);
+    code_end_offset = GRN_UINT32_VALUE_AT(&offsets, i * 2 + 1);
+    grn_table_group_keys_parse_one(ctx,
+                                   table,
+                                   keys_expression,
+                                   code_start_offset,
+                                   code_end_offset,
+                                   &(keys[i]));
+    if (ctx->rc != GRN_SUCCESS) {
+      goto exit;
+    }
+  }
+  *n_keys = n_offsets;
+
+exit :
+  if (ctx->rc != GRN_SUCCESS) {
+    if (keys) {
+      size_t i;
+      size_t n_offsets = GRN_UINT32_VECTOR_SIZE(&offsets) / 2;
+      for (i = 0; i < n_offsets; i++) {
+        if (keys[i].key) {
+          grn_obj_unref(ctx, keys[i].key);
+        }
+      }
+      GRN_FREE(keys);
+      keys = NULL;
+    }
+  }
+  if (keys_expression) {
+    grn_obj_close(ctx, keys_expression);
+  }
+  GRN_OBJ_FIN(ctx, &offsets);
+
+  GRN_API_RETURN(keys);
+}
